from __future__ import unicode_literals

import json
import psys
import six
import uuid

import mock
import pytest
import httmock

from collections import Sequence
from object_validator import String, Integer, DictScheme
from six import text_type as str

import pyscaleio
from pyscaleio import constants
from pyscaleio import exceptions

from pyscaleio import ScaleIOClient
from pyscaleio.manager import ScaleIOClientsManager
from pyscaleio.models import BaseResource, Sdc, StoragePool, System
from pyscaleio.models import Volume, ExportsInfo


@pytest.fixture
def client(request):

    client = ScaleIOClient.from_args("localhost", "admin", "passwd")
    pyscaleio.add_client(client)
    request.addfinalizer(ScaleIOClientsManager().deregister)
    return client


@pytest.fixture
def modelklass(request):

    def generate_klass(name, subclasses, dict):
        if six.PY2:
            name = psys.b(name)
        else:
            name = psys.u(name)
        return type(name, subclasses, dict)
    return generate_klass


@httmock.urlmatch(path=r".*login")
def login_payload(url, request):
    return httmock.response(200,
        json.dumps("some_random_token_string"),
        request=request
    )


def mock_resource_get(resource, resource_id, payload):
    path = r".*/api/instances/{0}::{1}".format(resource, resource_id)

    @httmock.urlmatch(path=path, method="get")
    def instance_of_payload(url, request):
        return httmock.response(200, payload, request=request)
    return instance_of_payload


def mock_resources_get(resource, payload):
    path = r".*/api/types/{0}/instances".format(resource)

    @httmock.urlmatch(path=path, method="get")
    def instances_of_payload(url, request):
        return httmock.response(200, payload, request=request)
    return instances_of_payload


def mock_volume(override=None):
    volume_model = {
        "id": str(uuid.uuid4()),
        "sizeInKb": (8 * constants.GIGABYTE) // constants.KILOBYTE,
        "storagePoolId": str(uuid.uuid4()),
        "useRmcache": False,
        "volumeType": constants.VOLUME_TYPE_THICK,
        "mappedSdcInfo": []
    }
    volume_model.update(override or {})
    return volume_model


def test_base_model_name(client):

    assert BaseResource.__resource__ is None
    with mock.patch("pyscaleio.models.BaseResource.__scheme__", {}):
        assert BaseResource._get_name() == "BaseResource"
        assert BaseResource(instance={})._get_name() == "BaseResource"

        with mock.patch(
            "pyscaleio.models.BaseResource.__resource__",
            "TestResourceName"
        ):
            assert BaseResource._get_name() == "TestResourceName"
            assert BaseResource(instance={})._get_name() == "TestResourceName"


@pytest.mark.parametrize(("name", "result"), [
    ("Volume", "Volume"), ("VTree", "VTree"),
    ("StoragePool", "StoragePool"), ("Sdc", "Sdc"),
])
def test_custom_model_name(client, modelklass, name, result):

    assert BaseResource.__resource__ is None
    with mock.patch("pyscaleio.models.BaseResource.__scheme__", {}):
        klass = modelklass(name, (BaseResource,), {"__scheme__": {}})
        assert klass._get_name() == result
        assert klass(instance={})._get_name() == result


@pytest.mark.parametrize("scheme", [
    {}, {"id": String()}
])
def test_base_model_scheme(client, modelklass, scheme):

    with mock.patch("pyscaleio.models.BaseResource.__scheme__", scheme):
        result = BaseResource._get_scheme()
        assert isinstance(result, DictScheme)
        assert result._DictScheme__scheme == scheme

        result = BaseResource(instance={"id": "test"})._get_scheme()
        assert isinstance(result, DictScheme)
        assert result._DictScheme__scheme == scheme


@pytest.mark.parametrize(("scheme", "data"), [
    ({"name": String()}, {"name": "test"}),
    (
        {"name": String(), "size": Integer()},
        {"name": "test", "size": 8 * constants.GIGABYTE}
    )
])
def test_custom_model_scheme(client, modelklass, scheme, data):

    base_scheme = {"id": String()}
    full_scheme = base_scheme
    full_scheme.update(scheme)

    data = data
    data.update({"id": "test_id"})

    with mock.patch(
        "pyscaleio.models.BaseResource.__scheme__",
        base_scheme
    ):
        klass = modelklass("Volume", (BaseResource,), {"__scheme__": scheme})
        result = klass._get_scheme()
        assert isinstance(result, DictScheme)
        assert result._DictScheme__scheme == full_scheme

        result = klass(instance=data)._get_scheme()
        assert isinstance(result, DictScheme)
        assert result._DictScheme__scheme == full_scheme


def test_model_initialize(client, modelklass):

    with mock.patch("pyscaleio.models.BaseResource.__scheme__", {}):
        klass = modelklass("Volume", (BaseResource,), {"__scheme__": {}})

        volume_id = "test"
        volume_payload = mock_resource_get(
            "Volume", volume_id, {"id": volume_id})
        with httmock.HTTMock(login_payload, volume_payload):
            assert klass() == {}
            assert len(klass()) == 0

            assert klass(instance_id=volume_id) == {"id": volume_id}

        assert klass(instance={"id": volume_id}) == {"id": volume_id}


def test_model_initialize_negative(client, modelklass):

    with mock.patch("pyscaleio.models.BaseResource.__scheme__", {}):
        klass = modelklass("Volume", (BaseResource,), {"__scheme__": {}})

        with pytest.raises(exceptions.ScaleIONotBothParameters):
            klass(instance_id="test", instance={"id": "test"})


def test_model_validation(client, modelklass):

    klass = modelklass("Volume", (BaseResource,), {
        "__scheme__": {
            "name": String()
        }
    })

    volume_payload = mock_resource_get("Volume", "test", {
        "id": "test",
        "name": "test_volume"
    })
    with httmock.HTTMock(login_payload, volume_payload):
        volume = klass("test")
        assert volume.get("id") == "test"
        assert volume.get("name") == "test_volume"


def test_model_validation_negative(client, modelklass):

    klass = modelklass("Volume", (BaseResource,), {
        "__scheme__": {
            "name": String()
        }
    })
    volume_payload = mock_resource_get("Volume", "test", {"id": "test"})
    with httmock.HTTMock(login_payload, volume_payload):
        with pytest.raises(exceptions.ScaleIOValidationError) as e:
            klass("test")
        assert "instance['name'] is missing" in str(e)


@pytest.mark.parametrize(("old_payload", "new_payload"), [
    (
        {"id": "test", "name": "test_volume"},
        {"id": "test", "name": "test_volume_changed"},
    ),
    (
        {"id": "test"},
        {"id": "test", "name": "test_volume"},
    ),
    (
        {"id": "test", "name": "test_volume"},
        {"id": "test"},
    )
])
def test_model_update(client, modelklass, old_payload, new_payload):

    klass = modelklass("Volume", (BaseResource,), {
        "__scheme__": {
            "name": String(optional=True)
        }
    })
    volume_payload = mock_resource_get("Volume", "test", old_payload)
    volume_update_payload = mock_resource_get("Volume", "test", new_payload)

    with httmock.HTTMock(login_payload, volume_payload):
        volume = klass("test")
        assert volume == old_payload

    with httmock.HTTMock(volume_update_payload):
        volume.update()
        assert volume == new_payload


def test_volume_model(client):

    volume_payload = mock_resource_get(Volume._get_name(), "test",
        mock_volume({"id": "test"})
    )
    system_payload = mock_resources_get(System._get_name(), [{
        "id": "system"
    }])
    with httmock.HTTMock(login_payload, volume_payload, system_payload):
        volume = Volume("test")

        assert volume.name is None
        assert volume.size == 8 * constants.GIGABYTE
        assert volume.type == constants.VOLUME_TYPE_THICK
        assert isinstance(volume.exports, Sequence)
        assert not volume.exports

        with mock.patch("pyscaleio.models.System.__scheme__", {}):
            assert volume.path == "/dev/disk/by-id/emc-vol-system-test"


def test_volume_model_exports(client):

    volume_exports = [{
        "sdcId": "sdc01",
        "sdcIp": "127.0.0.1",
        "limitIops": 0,
        "limitBwInMbps": 0
    }]
    volume_payload = mock_resource_get(Volume._get_name(), "test",
        mock_volume({
            "id": "test",
            "mappedSdcInfo": volume_exports
        })
    )
    sdc_payloads = [mock_resource_get(Sdc._get_name(), sdc_id, {
        "id": sdc_id,
        "sdcIp": sdc_id,
        "sdcGuid": str(uuid.uuid4()),
        "sdcApproved": True,
    }) for sdc_id in ("sdc0{0}".format(i) for i in range(1, 3))]

    with httmock.HTTMock(login_payload, volume_payload):
        volume = Volume("test")

        assert volume.exports
        assert isinstance(volume.exports, (Sequence, ExportsInfo))

    with httmock.HTTMock(*sdc_payloads):
        sdc1 = Sdc("sdc01")
        sdc2 = Sdc("sdc02")

        assert sdc1 in volume.exports
        assert sdc2 not in volume.exports
        assert "some_string" not in volume.exports


@pytest.mark.parametrize(("kw", "result"), [
    ({"sdc_id": "test"}, {"sdcId": "test"}),
    ({"sdc_guid": "test"}, {"guid": "test"}),
    (
        {"sdc_id": "test", "multiple": True},
        {"sdcId": "test", "allowMultipleMappings": "TRUE"}
    ),
])
def test_volume_export(client, kw, result):

    with mock.patch("pyscaleio.models.Volume.__scheme__", {}):
        volume = Volume(instance={"id": "test", "links": []})

    with mock.patch("pyscaleio.ScaleIOClient.perform_action_on") as m:
        volume.export(**kw)
        m.assert_called_once_with("Volume", "test", "addMappedSdc", result)


@pytest.mark.parametrize(("kw", "result"), [
    ({"sdc_id": "test"}, {"sdcId": "test"}),
    ({"sdc_guid": "test"}, {"guid": "test"}),
    ({}, {"allSdcs": ""})
])
def test_volume_unexport(client, kw, result):

    with mock.patch("pyscaleio.models.Volume.__scheme__", {}):
        volume = Volume(instance={"id": "test", "links": []})

    with mock.patch("pyscaleio.ScaleIOClient.perform_action_on") as m:
        volume.unexport(**kw)
        m.assert_called_once_with("Volume", "test", "removeMappedSdc", result)


@pytest.mark.parametrize("method", ["export", "unexport"])
def test_volume_unexport_negative(client, method):

    with mock.patch("pyscaleio.models.Volume.__scheme__", {}):
        volume = Volume(instance={"id": "test", "links": []})

    with mock.patch("pyscaleio.ScaleIOClient.perform_action_on") as m:
        with pytest.raises(exceptions.ScaleIONotBothParameters):
            getattr(volume, method)(**{"sdc_id": "test", "sdc_guid": "test"})
        m.assert_not_called()


@pytest.mark.parametrize(("kw", "result"), [
    ({"sdc_id": "test", "iops": 1000}, {"sdcId": "test", "iopsLimit": "1000"}),
    ({"sdc_id": "test", "iops": 0}, {"sdcId": "test", "iopsLimit": "0"}),
    ({"sdc_guid": "test", "mbps": 2048}, {"guid": "test", "bandwidthLimitInKbps": "2097152"}),
    ({"sdc_guid": "test", "mbps": 0}, {"guid": "test", "bandwidthLimitInKbps": "0"})
])
def test_volume_throttle_positive(client, kw, result):

    with mock.patch("pyscaleio.models.Volume.__scheme__", {}):
        volume = Volume(instance={"id": "test", "links": []})

    with mock.patch("pyscaleio.ScaleIOClient.perform_action_on") as m:
        volume.throttle(**kw)
        m.assert_called_once_with("Volume", "test", "setMappedSdcLimits", result)


@pytest.mark.parametrize(("kw", "exception"), [
    ({"sdc_id": "test"}, exceptions.ScaleIORequiredParameters),
    ({"sdc_id": "test", "iops": 5}, exceptions.ScaleIOInvalidLimit),
    ({"sdc_id": "test", "mbps": 1024.5}, exceptions.ScaleIOInvalidLimit),
    ({"sdc_id": "test", "sdc_guid": "test_guid"}, exceptions.ScaleIONotBothParameters),
])
def test_volume_throttle_negative(client, kw, exception):

    with mock.patch("pyscaleio.models.Volume.__scheme__", {}):
        volume = Volume(instance={"id": "test", "links": []})

    with mock.patch("pyscaleio.ScaleIOClient.perform_action_on") as m:
        with pytest.raises(exception):
            volume.throttle(**kw)
        m.assert_not_called()


@pytest.mark.parametrize(("kw", "result"), [
    ({}, {}), ({"name": "test_snapshot"}, {"snapshotName": "test_snapshot"})
])
def test_volume_snapshot(client, kw, result):
    volume_data = {
        "id": "base_volume",
        "links": [],
        "volumeType": constants.VOLUME_TYPE_THIN,
        "useRmcache": False,
        "sizeInKb": 1048576,
        "storagePoolId": "pool_id"
    }
    volume = Volume(instance=volume_data)

    snapshot_data = volume_data.copy()
    snapshot_data.update({
        "id": "volume_snapshot",
        "volumeType": constants.VOLUME_TYPE_SNAPSHOT,
        "ancestorVolume": "base_volume",
    })
    snapshot_payload = mock_resource_get(Volume._get_name(),
        "volume_snapshot", snapshot_data)

    system_payload = mock_resources_get(System._get_name(), [{
        "id": "test", "restrictedSdcModeEnabled": True
    }])
    method_data = {
        "snapshotDefs": [dict({"volumeId": "base_volume"}, **result)]
    }
    with httmock.HTTMock(login_payload, system_payload, snapshot_payload):
        with mock.patch(
            "pyscaleio.ScaleIOClient.perform_action_on",
            side_effect=[{"volumeIdList": ["volume_snapshot"]}]
        ) as m:
            snapshot = volume.snapshot(**kw)
            m.assert_called_once_with("System", "test", "snapshotVolumes", method_data)

        assert isinstance(snapshot, Volume)
        assert snapshot.type == constants.VOLUME_TYPE_SNAPSHOT
        assert snapshot.get("ancestorVolume") == "base_volume"


@pytest.mark.parametrize(("kw", "result"), [
    ({"name": "test"}, {"name": "test"}),
    ({"rmcache": True}, {"useRmcache": True}),
    ({"thin": False}, {"volumeType": constants.VOLUME_TYPE_THICK}),
])
def test_volume_create(client, kw, result):

    args = (1, "test_pool")
    full_result = {
        "volumeSizeInKb": str(1048576),
        "storagePoolId": "test_pool",
        "volumeType": constants.VOLUME_TYPE_THIN,
    }
    full_result.update(result)

    with mock.patch("pyscaleio.models.MutableResource.create") as m:
        Volume.create(*args, **kw)
        m.assert_called_once_with(full_result)


def test_volume_one_by_name(client):

    volume_id = "test_id"
    volume_name = "test_name"

    volume_payload = mock_resource_get(Volume._get_name(), volume_id, {
        "id": volume_id,
        "name": volume_name
    })

    call_args = (Volume._get_name(), "queryIdByKey", {"name": volume_name})
    with mock.patch("pyscaleio.models.Volume.__scheme__", {}):
        with mock.patch(
            "pyscaleio.ScaleIOClient.perform_action_on_type",
            side_effect=[volume_id]
        ) as m:
            with httmock.HTTMock(login_payload, volume_payload):
                volume = Volume.one_by_name(volume_name)
            m.assert_called_once_with(*call_args)

            assert isinstance(volume, Volume)
            assert volume.name == volume_name
            assert volume["id"] == volume_id


@pytest.mark.parametrize(("kw", "result"), [
    ({"checksum": True}, {"checksumEnabled": "TRUE"}),
    ({"rfcache": True}, {"useRfcache": "TRUE"}),
    (
        {"checksum": True, "rfcache": True},
        {"checksumEnabled": "TRUE", "useRfcache": "TRUE"}
    ),
    ({"name": "test_pool"}, {"name": "test_pool"})
])
def test_storage_pool_create(client, kw, result):

    args = ("domain_id",)
    full_result = {
        "protectionDomainId": "domain_id",
        "checksumEnabled": "FALSE",
        "useRfcache": "FALSE",
    }
    full_result.update(result)

    with mock.patch("pyscaleio.models.MutableResource.create") as m:
        StoragePool.create(*args, **kw)
        m.assert_called_once_with(full_result)


def test_storage_pool_one_by_name(client):

    pool_id = "test_id"
    pool_name = "test_name"
    domain_id = "domain_id"
    domain_name = "domain_name"

    pool_payload = mock_resource_get(StoragePool._get_name(), pool_id, {
        "id": pool_id,
        "name": pool_name,
        "protectionDomainId": domain_id,
        "checksumEnabled": False,
        "useRfcache": False,
    })

    call_args = (StoragePool._get_name(), "queryIdByKey", {
        "name": pool_name,
        "protectionDomainName": domain_name,
    })
    with mock.patch(
        "pyscaleio.ScaleIOClient.perform_action_on_type",
        side_effect=[pool_id]
    ) as m:
        with httmock.HTTMock(login_payload, pool_payload):
            pool = StoragePool.one_by_name(pool_name, domain_name)
        m.assert_called_once_with(*call_args)

        assert isinstance(pool, StoragePool)
        assert pool.name == pool_name
        assert pool["id"] == pool_id
        assert pool.checksum_enabled is False
        assert pool.rfcache_enabled is False


def test_sdc_one_by_ip(client):

    sdc_id = "test_id"
    sdc_ip = "172.20.34.126"
    sdc_guid = str(uuid.uuid4()).upper()

    sdc_payload = mock_resource_get(Sdc._get_name(), sdc_id, {
        "id": sdc_id,
        "name": "test_sdc",
        "sdcIp": sdc_ip,
        "sdcGuid": sdc_guid,
        "sdcApproved": True,
    })

    call_args = (Sdc._get_name(), "queryIdByKey", {"ip": sdc_ip})
    with mock.patch(
        "pyscaleio.ScaleIOClient.perform_action_on_type",
        side_effect=[sdc_id]
    ) as m:
        with httmock.HTTMock(login_payload, sdc_payload):
            sdc = Sdc.one_by_ip(sdc_ip)
        m.assert_called_once_with(*call_args)

        assert isinstance(sdc, Sdc)
        assert sdc["id"] == sdc_id
        assert sdc.name == "test_sdc"
        assert sdc.ip == sdc_ip
        assert sdc.guid == sdc_guid
        assert sdc.is_approved



from __future__ import unicode_literals

import os
from six import text_type as str
from collections import Mapping, Sequence

from inflection import camelize, underscore
from object_validator import validate, ValidationError
from object_validator import DictScheme, List, String, Integer, Bool

import pyscaleio
from pyscaleio import config
from pyscaleio import constants
from pyscaleio import exceptions
from pyscaleio import utils


class BaseResource(Mapping):
    """Base resource model."""

    __scheme__ = {
        "id": String(),
        "links": List(
            DictScheme({
                "href": String(),
                "rel": String()
            }), optional=True
        )
    }
    """Data scheme for instance validation."""

    __parents__ = None
    """
    References to parent resources by fields.

    Example:
        frozenset([
            ("parentField", "ResourceClassName")
        ])
    """

    __resource__ = None
    """
    Custom resource name that overrides default
    name based on name of resource class.
    """

    @classmethod
    def _get_name(cls):
        """Returns resource name.

        Attention: for internal use only!
        """

        resource = cls.__resource__
        if not resource:
            resource = camelize(underscore(cls.__name__))
        return resource

    @classmethod
    def _get_scheme(cls):
        """Returns resource scheme for proper validation.

        Attention: for internal use only!
        """

        scheme = {}
        for base in cls.mro():
            if bool(
                issubclass(base, BaseResource) and
                getattr(base, "__scheme__", None)
            ):
                scheme.update(base.__scheme__)

        return DictScheme(scheme, ignore_unknown=True)

    @classmethod
    def one(cls, instance_id, **kwargs):
        """Returns instance of resource.

        :param instance_id: id of resource instance

        :returns: instance of resource
        """

        return cls(instance_id, **kwargs)

    @pyscaleio.inject
    @classmethod
    def all(cls, client, instance_ids=None, **kwargs):
        """Returns list of resource instances.

        :param instance_ids: list of instance ids (optional)

        :returns: list of resource instances
        """

        if not instance_ids:
            instances = client.get_instances_of(cls._get_name())
        else:
            if isinstance(instance_ids, basestring):
                instance_ids = (instance_ids,)
            instances = client.perform_action_on_type(
                cls._get_name(), "queryBySelectedIds", {"ids": instance_ids})

        return [cls(instance=instance, client=client)
            for instance in instances
        ]

    @pyscaleio.inject
    def __init__(self, client, instance_id=None, instance=None):
        self._client = client
        self._scheme = {}

        if instance_id and instance:
            raise exceptions.ScaleIONotBothParameters("instance_id", "instance")

        if instance_id:
            instance = self._client.get_instance_of(self._get_name(), instance_id)

        self._instance = self._validate(instance or {})

    def __getitem__(self, key):
        return self._instance[key]

    def __iter__(self):
        return iter(self._instance)

    def __len__(self):
        return len(self._instance)

    @property
    def links(self):
        return self["links"]

    def _validate(self, instance):
        """Validates the instance if resource according to scheme.

        Attention: for internal use only!
        """

        try:
            return validate("instance", instance, self._get_scheme())
        except ValidationError as e:
            raise exceptions.ScaleIOValidationError(e)

    def update(self):
        """Updates resource instance."""

        instance = self._client.get_instance_of(self._get_name(), self["id"])
        instance = self._validate(instance)

        fields = set(list(instance) + list(self._instance))
        for field in fields:
            try:
                instance[field]
            except KeyError:
                del self._instance[field]
            else:
                self._instance[field] = instance[field]


class EditableResource(BaseResource):
    """Resource model with editable properties."""

    def perform(self, action, data):
        """Performs action on resource instance.

        :param action: action name
        :param data: action data payload
        """

        return self._client.perform_action_on(self._get_name(), self["id"], action, data)


class MutableResource(EditableResource):
    """Resource model that can be created/deleted."""

    @pyscaleio.inject
    @classmethod
    def create(cls, client, instance, **kwargs):
        """Created instance of resource.

        :param instance: instance payload
        """

        instance_id = client.create_instance_of(cls._get_name(), instance)

        return cls(instance_id, **kwargs)

    def delete(self, data=None):
        """Deletes instance of resource.

        :param data: action data payload (optional)
        """

        self.perform("remove{0}".format(self._get_name()), data or {})


class System(EditableResource):
    """System resource model."""

    __scheme__ = {
        "name": String(optional=True),
        "restrictedSdcModeEnabled": Bool(),
    }

    @property
    def name(self):
        return self.get("name")

    @property
    def is_restricted(self):
        return self["restrictedSdcModeEnabled"]


class ProtectionDomain(MutableResource):
    """ProtectionDomain resource model."""

    __scheme__ = {
        "name": String(),
        "systemId": String(),
    }
    __parents__ = frozenset([
        ("systemId", "System")
    ])

    @property
    def name(self):
        return self["name"]


class StoragePool(MutableResource):
    """StoragePool resource model."""

    __scheme__ = {
        "name": String(optional=True),
        "protectionDomainId": String(),
        "checksumEnabled": Bool(),
        "useRfcache": Bool(),
    }
    __parents__ = frozenset([
        ("protectionDomainId", "ProtectionDomain")
    ])

    @pyscaleio.inject
    @classmethod
    def one_by_name(cls, client, name, domain_name):
        """Returns StoragePool instance by name and protection domain name.

        :param name: storage pool name
        :param domain_name: protection domain name

        :rtype: pyscaleio.StoragePool
        """

        data = {
            "name": name,
            "protectionDomainName": domain_name
        }
        pool_id = client.perform_action_on_type(
            cls._get_name(), "queryIdByKey", data)

        return cls(pool_id, client=client)

    @pyscaleio.inject
    @classmethod
    def create(cls, client, domain, checksum=False, rfcache=False, name=None, **kwargs):
        """Creates StoragePool instance.

        :param domain: protection domain id (required)
        :param checksum: enable checksum protection (default is False)
        :param rfcache: use read-only flash cache (default is False)

        :rtype: pyscaleio.StoragePool
        """

        pool = {
            "protectionDomainId": domain,
            "checksumEnabled": utils.bool_to_str(checksum),
            "useRfcache": utils.bool_to_str(rfcache),
        }
        if name:
            pool["name"] = name

        return super(StoragePool, cls).create(pool, **kwargs)

    @property
    def name(self):
        return self.get("name")

    @property
    def checksum_enabled(self):
        return self["checksumEnabled"]

    @property
    def rfcache_enabled(self):
        return self["useRfcache"]

    def create_volume(self, size, **kwargs):
        """Creates Volume instance in current StoragePool.

        :param size: volume size in GB (required)

        :rtype: pyscaleio.Volume
        """

        kwargs["client"] = self._client
        return Volume.create(size, self["id"], **kwargs)


class VTree(BaseResource):
    """Volume Tree (VTree) resource model."""

    __scheme__ = {
        "name": String(optional=True),
        "baseVolumeId": String(),
        "storagePoolId": String(),
    }

    __parents__ = frozenset([
        ("baseVolumeId", "Volume"),
        ("storagePoolId", "StoragePool")
    ])


class Sdc(MutableResource):
    """SDC resource model."""

    __scheme__ = {
        "name": String(optional=True),
        "sdcIp": String(),
        "sdcGuid": String(),
        "sdcApproved": Bool(),
    }
    __parents__ = frozenset([
        ("systemId", "System")
    ])

    @pyscaleio.inject
    @classmethod
    def all_approved(cls, client, **kwargs):
        """Returns list of all approved SDCs."""

        instances = client.perform_action_on_type(
            cls._get_name(), "queryAllApprovedSdc", {})

        return [cls(instance=instance, client=client)
            for instance in instances
        ]

    @pyscaleio.inject
    @classmethod
    def one_by_ip(cls, client, ip_address, **kwargs):
        """Returns SDC instance by specified IP address.

        :param ip_address: IP address of SDC

        :rtype: pyscaleio.SDC
        """

        instance_id = client.perform_action_on_type(
            cls._get_name(), "queryIdByKey", {"ip": ip_address})

        return cls(instance_id, client=client)

    @property
    def name(self):
        return self.get("name")

    @property
    def ip(self):
        return self["sdcIp"]

    @property
    def guid(self):
        return self["sdcGuid"]

    @property
    def is_approved(self):
        return self["sdcApproved"]


class ExportsInfo(Sequence):
    """Information about volume exports."""

    __scheme__ = List(DictScheme({
        "sdcId": String(),
        "sdcIp": String(),
        "limitIops": Integer(),
        "limitBwInMbps": Integer()
    }), optional=True)

    def __init__(self, data=None):
        self._data = data or []

    def __getitem__(self, index):
        return self._data[index]

    def __len__(self):
        return len(self._data)

    def __contains__(self, key):
        if isinstance(key, Sdc):
            return key["id"] in (e["sdcId"] for e in self._data)
        else:
            super(ExportsInfo, self).__contains__(key)


class Volume(MutableResource):
    """Volume resource model."""

    __scheme__ = {
        "name": String(optional=True),
        "mappedSdcInfo": ExportsInfo.__scheme__,
        "useRmcache": Bool(),
        "sizeInKb": Integer(),
        "storagePoolId": String(),
        "volumeType": String(choices=constants.VOLUME_TYPES),
        "ancestorVolumeId": String(optional=True),
    }
    __parents__ = frozenset([
        ("ancestorVolumeId", "Volume"),
        ("storagePoolId", "StoragePool"),
        ("vtreeId", "VTree"),
    ])

    @pyscaleio.inject
    @classmethod
    def one_by_name(cls, client, name, **kwargs):
        """Returns volume instance by name.

        :param name: volume name (required)

        :rtype: pyscaleio.Volume
        """

        volume_id = client.perform_action_on_type(
            cls._get_name(), "queryIdByKey", {"name": name})

        return cls(volume_id, **kwargs)

    @classmethod
    def create(cls, size, pool, name=None, rmcache=None, thin=True, **kwargs):
        """Creates Volume instance.

        :param size: volume size in GB (required)
        :param pool: storage pool id (required)
        :param name: volume name
        :param thin: is volume 'thin' or 'thick' provisioned
        :param rmcache: volume rmcache

        :rtype: pyscaleio.Volume
        """

        volume_size = (size * constants.GIGABYTE) // constants.KILOBYTE
        volume = {
            "volumeSizeInKb": str(volume_size),
            "storagePoolId": pool
        }
        if name:
            volume["name"] = name
        if rmcache:
            volume["useRmcache"] = rmcache

        volume["volumeType"] = constants.VOLUME_TYPE_THICK
        if thin:
            volume["volumeType"] = constants.VOLUME_TYPE_THIN

        return super(Volume, cls).create(volume, **kwargs)

    @property
    def name(self):
        return self.get("name")

    @property
    def size(self):
        return self["sizeInKb"] * constants.KILOBYTE

    @property
    def type(self):
        return self.get("volumeType")

    @property
    def exports(self):
        return ExportsInfo(self.get("mappedSdcInfo"))

    @property
    def path(self):
        device_name = config.VOLUME_NAME.format(
            system_id=self._client.system["id"],
            volume_id=self["id"]
        )
        return os.path.join(config.VOLUME_PREFIX, device_name)

    def rename(self, name):
        """Changes volume name.

        :param name: new volume name
        """

        return super(Volume, self).perform("setVolumeName", {"newName": name})

    def resize(self, size):
        """Changes volumes size.

        :param size: new volume size in GB (required)
        """

        return super(Volume, self).perform("setVolumeSize", {"sizeInGB": str(size)})

    def snapshot(self, name=None):
        """Creates snapshot of current volume.

        :param name: snapshot name
        """

        snapshot = {"volumeId": self["id"]}
        if name:
            snapshot["snapshotName"] = name

        result = self._client.system.perform(
            "snapshotVolumes", {"snapshotDefs": [snapshot]})

        return Volume(result["volumeIdList"][0])

    def throttle(self, sdc_id=None, sdc_guid=None, iops=None, mbps=None):
        """Throttles I/O on current volume.

        :param iops: I/O operations in seconds (> 10)
        :param mbps: network bandwidth as megabytes in seconds
        """

        if sdc_id and sdc_guid:
            raise exceptions.ScaleIONotBothParameters("sdc_id", "sdc_guid")

        if iops is None and mbps is None:
            raise exceptions.ScaleIORequiredParameters("iops", "mbps")

        if iops is not None and iops != 0 and iops <= 10:
            raise exceptions.ScaleIOInvalidLimit("iops", "must be greater than 10.")

        if mbps is not None:
            mbps = mbps * constants.KILOBYTE
            if bool(mbps % constants.KILOBYTE):
                raise exceptions.ScaleIOInvalidLimit("mbps", "must be granular to 1024 Kbps")

        data = {}
        if sdc_id:
            data["sdcId"] = sdc_id
        if sdc_guid:
            data["guid"] = sdc_guid
        if iops is not None:
            data["iopsLimit"] = str(iops)
        if mbps is not None:
            data["bandwidthLimitInKbps"] = str(mbps)

        return super(Volume, self).perform("setMappedSdcLimits", data)

    def export(self, sdc_id=None, sdc_guid=None, multiple=False):
        """Exports volume to specified SDC.

        :param sdc_id: id of SDC instance
        :param sdc_guid: guid of SDC instance
        :param multiple: allows export to multiple SDCs (optional)
        """

        if sdc_id and sdc_guid:
            raise exceptions.ScaleIONotBothParameters("sdc_id", "sdc_guid")

        data = {}
        if sdc_id:
            data["sdcId"] = sdc_id
        if sdc_guid:
            data["guid"] = sdc_guid

        if multiple:
            data["allowMultipleMappings"] = utils.bool_to_str(multiple)

        return super(Volume, self).perform("addMappedSdc", data)

    def unexport(self, sdc_id=None, sdc_guid=None):
        """Unexports volume from specified SDC.
        Without parameters unexports volume from all SDCs

        :param sdc_id: id of SDC instance
        :param sdc_guid: guid if SDC instance
        """

        if sdc_id and sdc_guid:
            raise exceptions.ScaleIONotBothParameters("sdc_id", "sdc_guid")

        data = {}
        if sdc_id:
            data["sdcId"] = sdc_id
        elif sdc_guid:
            data["guid"] = sdc_guid
        else:
            data["allSdcs"] = ""

        return super(Volume, self).perform("removeMappedSdc", data)

    def delete(self, mode=constants.VOLUME_REMOVE_ONLY_ME):
        """Removes volume with specified mode.

        :param mode: volume remove mode (optional)
        """

        return super(Volume, self).delete({"removeMode": mode})



from psys import Error


class ScaleIOError(Error):
    def __init__(self, code, message, error_code=None):
        super(ScaleIOError, self).__init__(
            "ScaleIOError: code={0}, message={1}", code, message
        )
        self.status_code = code
        self.error_code = error_code or 0


class ScaleIOAuthError(ScaleIOError):
    def __init__(self):
        super(ScaleIOAuthError, self).__init__(401, "Unauthorized")


class ScaleIOMalformedError(ScaleIOError):
    def __init__(self):
        super(ScaleIOMalformedError, self).__init__(500, "Malformed response")


class ScaleIOInvalidClient(Error):
    def __init__(self):
        super(ScaleIOInvalidClient, self).__init__("Invalid ScaleIO client instance.")


class ScaleIOClientAlreadyRegistered(Error):
    def __init__(self, key):
        super(ScaleIOClientAlreadyRegistered, self).__init__(
            "ScaleIOClient with key '{0}' already registered.", key)


class ScaleIOClientNotRegistered(Error):
    def __init__(self, key):
        super(ScaleIOClientNotRegistered, self).__init__(
            "ScaleIOClient with key '{0}' not registered.", key)


class ScaleIOEmptyClientRegistry(Error):
    def __init__(self):
        super(ScaleIOEmptyClientRegistry, self).__init__(
            "Clients for ScaleIO not registered.")


class ScaleIOValidationError(Error):
    def __init__(self, exc):
        super(ScaleIOValidationError, self).__init__(
            "Resource validation error: {0}", exc)


class ScaleIOConfigError(Error):
    def __init__(self, exc):
        super(ScaleIOConfigError, self).__init__(
            "Config validation error: {0}", exc)


class ScaleIOInvalidParameters(Error):
    def __init__(self, *args, **kwargs):
        super(ScaleIOInvalidParameters, self).__init__(*args, **kwargs)


class ScaleIONotBothParameters(ScaleIOInvalidParameters):
    def __init__(self, first, second):
        super(ScaleIONotBothParameters, self).__init__(
            "Use either '{0}' or '{1}', not both", first, second)


class ScaleIORequiredParameters(ScaleIOInvalidParameters):
    def __init__(self, *params):
        super(ScaleIORequiredParameters, self).__init__(
            "One of the parameter(s) must be specified: {0}", ", ".join(params))


class ScaleIOInvalidLimit(ScaleIOInvalidParameters):
    def __init__(self, limit, message):
        super(ScaleIOInvalidLimit, self).__init__(
            "Invalid '{0}' limit: {1}", limit, message)



from __future__ import unicode_literals

from collections import MutableMapping, MutableSequence
from functools import wraps


class singleton(type):
    """Singleton meta-class."""

    _instances = {}

    def __call__(cls, *args, **kwargs):
        if cls not in cls._instances:
            cls._instances[cls] = super(singleton, cls).__call__(*args, **kwargs)
        return cls._instances[cls]


def _drop_none(mapping):
    """
    Removes all keys that points to None values.

    >>> _drop_none({})
    {}
    >>> _drop_none({'a': 'A', 'b': 'B'}) == {'a': 'A', 'b': 'B'}
    True
    >>> _drop_none({'a': 'A', 'b': None}) == {'a': 'A'}
    True
    >>> _drop_none({'a': 'A', 'b': None, 'c': 'C', 'd': None}) == {'a': 'A', 'c': 'C'}
    True
    >>> _drop_none(('a', 'b'))
    Traceback (most recent call last):
    ...
    AssertionError: assert isinstance(('a', 'b'), MutableMapping)
    """

    assert isinstance(mapping, MutableMapping)
    for key in mapping.copy():
        if mapping[key] is None:
            mapping.pop(key)

    return mapping


def drop_none(func):
    """
    >>> drop_none(lambda: {'a': 'A', 'b': None})() == {'a': 'A'}
    True
    >>> drop_none(lambda: [{'a': 'A', 'b': None}])() == [{'a': 'A'}]
    True
    >>> drop_none(lambda: "test")() == "test"
    True
    """

    @wraps(func)
    def wrapper(*args, **kwargs):
        results = func(*args, **kwargs)
        if isinstance(results, MutableSequence):
            return [_drop_none(result) for result in results]
        elif isinstance(results, MutableMapping):
            return _drop_none(results)
        else:
            return results
    return wrapper


def bool_to_str(value):
    """
    Converts bool value to string.

    >>> bool_to_str(True) == "TRUE"
    True
    >>> bool_to_str(False) == "FALSE"
    True
    >>> bool_to_str("test")
    Traceback (most recent call last):
    ...
    AssertionError: assert isinstance('test', bool)
    """

    assert isinstance(value, bool)

    return "TRUE" if value is True else "FALSE"


class _decorator(object):
    """Adopts decorator function to methods and functions."""

    def __init__(self, decorator, func):
        self.decorator = decorator
        self.func = func

    def __call__(self, *args, **kwargs):
        return self.decorator(self.func)(*args, **kwargs)

    def __get__(self, instance, owner):
        return self.decorator(self.func.__get__(instance, owner))


def decorator(decorator):
    def wrapper(func):
        return _decorator(decorator, func)
    return wrapper



from __future__ import unicode_literals

from six import add_metaclass

from pyscaleio import exceptions
from pyscaleio import utils
from pyscaleio.client import ScaleIOClient


@add_metaclass(utils.singleton)
class ScaleIOClientsManager(object):
    """ScaleIO Clients manager."""

    def __init__(self):
        self.__registry_cleanup()

    @property
    def clients(self):
        return self.__clients

    @property
    def default(self):
        return self.__default

    @staticmethod
    def _construct_key(instance):
        """Returns key for ScaleIOClient instance."""

        # TODO: combine key with port if exists
        return instance.session.host

    def __registry_cleanup(self):
        """Cleanups the clients registry."""

        self.__clients = {}
        self.__default = None

    def register(self, instance, default=True):
        """Registers ScaleIOClient instances."""

        if not isinstance(instance, ScaleIOClient):
            raise exceptions.ScaleIOInvalidClient()

        instance_key = self._construct_key(instance)
        if instance_key in self.__clients:
            raise exceptions.ScaleIOClientAlreadyRegistered(instance_key)

        self.__clients[instance_key] = instance
        if not self.__default:
            self.__default = instance

    def get_client(self, instance_key=None):
        """Returns ScaleIOClient instance."""

        if instance_key:
            try:
                return self.__clients[instance_key]
            except KeyError:
                raise exceptions.ScaleIOClientNotRegistered(instance_key)

        if self.__default:
            return self.__default

        raise exceptions.ScaleIOEmptyClientRegistry()

    def deregister(self, instance_key=None):
        """Deregisters ScaleIOClient instances."""

        if instance_key:
            client = self.__clients.pop(instance_key, None)
            if client is self.__default:
                self.__default = None
        else:
            self.__registry_cleanup()



from __future__ import unicode_literals

import json
import logging
import psys
import requests
import uuid

from functools import wraps
from six import text_type as str
from six.moves.urllib.parse import urljoin

import pyscaleio
from pyscaleio import config
from pyscaleio import exceptions
from pyscaleio import utils

try:
    from requests.packages import urllib3
except ImportError:
    import urllib3

urllib3.disable_warnings()
"""Disable certificate warnings."""

__api_version__ = 2.0
"""ScaleIO API Version."""

log = logging.getLogger(__name__)
"""Logger instance."""


class ScaleIOSession(object):
    """ScaleIO session base class."""

    __session = None
    """Session instance."""

    __endpoint = "{scheme}://{host}/api/"
    """Endpoint template."""

    def __init__(self, host, user, passwd, is_secure=True,
                 retries=config.REQUEST_RETRIES,
                 timeout=config.NETWORK_TIMEOUT):
        self.host = host
        self.scheme = "https" if is_secure else "http"

        self.user = user
        self.passwd = passwd

        self.timeout = timeout
        self.retries = retries

        self.token = None
        self.headers = {
            "Content-Type": "application/json",
            "Accept": "application/json; version={version}".format(
                version=__api_version__
            )
        }
        self.__session = requests.Session()
        self.__session.headers.update(self.headers)

    @property
    def endpoint(self):
        return self.__endpoint.format(
            scheme=self.scheme, host=self.host)

    def __expired(self):
        """Handle session expiring."""

        self.token = None
        self.__session.auth = None

        self.login()

    def __error(self, exc, request_uuid):
        """Handle request error."""

        error = self.__response(exc.response, request_uuid)
        raise exceptions.ScaleIOError(error["httpStatusCode"],
                                      error["message"],
                                      error["errorCode"])

    def __response(self, response, request_uuid):
        """Handle response object."""

        response_payload = response.text

        log.debug("ScaleIO response (%s): %s", request_uuid, response_payload)
        try:
            return json.loads(psys.u(response_payload))
        except (ValueError, TypeError):
            raise exceptions.ScaleIOMalformedError()

    def login(self, timeout=None):
        """Logins to ScaleIO REST Gateway."""

        url = urljoin(self.endpoint, "login")
        auth = (self.user, self.passwd)

        request_uuid = str(uuid.uuid4())
        log.debug("ScaleIO login (%s): url=%s, auth=%s, timeout=%s",
            request_uuid, url, auth, timeout)

        response = self.__session.get(
            url=url, auth=auth, allow_redirects=False,
            verify=False, timeout=timeout or self.timeout)
        try:
            response.raise_for_status()
        except requests.HTTPError as e:
            if e.response.status_code == 401:
                raise exceptions.ScaleIOAuthError()
            else:
                raise

        self.token = self.__response(response, request_uuid)
        self.__session.auth = (self.user, self.token)

    def logout(self, timeout=None):
        """Logout from ScaleIO REST Gateway and invalidates token."""

        if self.__session and self.token:
            self.__session.get(
                url=urljoin(self.endpoint, "logout"),
                allow_redirects=False, verify=False,
                timeout=timeout or self.timeout)

        self.token = None
        self.__session.auth = None

    def _send_request(self, method, url, params=None, data=None, headers=None):
        """Base method for sending requests."""

        headers = headers or {}
        retries = self.retries

        if not self.token:
            self.login()

        request_uuid = str(uuid.uuid4())
        log.debug("ScaleIO request (%s): method=%s, url=%s, params=%s, data=%s",
            request_uuid, method, url, params, data)

        response = None
        while retries > 0:
            response = self.__session.request(
                method=method,
                url=url,
                params=params,
                data=data,
                timeout=self.timeout,
                allow_redirects=False,
                headers=headers,
                verify=False,
            )
            try:
                response.raise_for_status()
            except requests.HTTPError as e:
                if e.response.status_code == 401:
                    self.__expired()
                    retries -= 1
                    continue
                else:
                    self.__error(e, request_uuid)
            else:
                if not response:
                    retries -= 1
                    continue
                else:
                    return self.__response(response, request_uuid)

    def get(self, path, params=None):
        return self._send_request(method="get",
            url=urljoin(self.endpoint, path), params=params)

    def post(self, path, data):
        return self._send_request(method="post",
            url=urljoin(self.endpoint, path), data=data)


class ScaleIOClient(object):
    """API Client for ScaleIO."""

    _session = None
    """ScaleIO session instance."""

    @classmethod
    def from_args(cls, *args, **kwargs):
        """Initialize from ScaleIOSession arguments."""

        return cls(session=ScaleIOSession(*args, **kwargs))

    def __init__(self, session):
        if not isinstance(session, ScaleIOSession):
            raise psys.Error(
                "ScaleIOClient must be initialized with ScaleIOSession.")
        self._session = session
        self._system = None

    @property
    def session(self):
        return self._session

    @property
    def system(self):
        from pyscaleio.models import System

        if not self._system:
            self._system = System.all(client=self)[0]
        return self._system

    def get_version(self):
        """Returns ScaleIO REST API version."""

        return self._session.get("version")

    def get_all_instances(self):
        """Returns all exists instances of all types."""

        return self._session.get("instances")

    @utils.drop_none
    def get_instances_of(self, resourse, params=None):
        """Returns list of instances of specified resource."""

        return self._session.get("types/{type}/instances".format(
            type=resourse), params=params
        )

    @utils.drop_none
    def get_instance_of(self, resourse, resourse_id):
        """Returns instance of specified resource type by id."""

        return self._session.get("instances/{type}::{id}".format(
            type=resourse, id=resourse_id)
        )

    def create_instance_of(self, resource, resource_data):
        """Creates instance of specified resource."""

        response = self._session.post("types/{type}/instances".format(
            type=resource), data=psys.u(json.dumps(resource_data))
        )
        return response["id"]

    def perform_action_on(self, resource, resource_id, action, action_data):
        """Performs action on single instance of specified resource type."""

        return self._session.post("instances/{type}::{id}/action/{action}".format(
            type=resource, id=resource_id, action=action),
            data=psys.u(json.dumps(action_data))
        )

    @utils.drop_none
    def perform_action_on_type(self, resource, action, action_data):
        """Performs action on specified resource type."""

        return self._session.post("types/{type}/instances/action/{action}".format(
            type=resource, action=action), data=psys.u(json.dumps(action_data))
        )


def _get_client(kwargs):
    """
    Returns ScaleIOClient instance by specified kwargs.
    Attention: for internal use only!
    """

    host, client = kwargs.pop("host", None), kwargs.pop("client", None)
    if host and client:
        raise exceptions.ScaleIONotBothParameters("host", "client")

    if client:
        if not isinstance(client, ScaleIOClient):
            raise exceptions.ScaleIOInvalidClient()
        return client
    else:
        return pyscaleio.get_client(host)


@utils.decorator
def inject(function):
    """
    Decorates and injects ScaleIOClient instance
    into decorated method or function.
    """

    @wraps(function)
    def wrapper(*args, **kwargs):
        client = _get_client(kwargs)
        return function(client, *args, **kwargs)
    return wrapper



from __future__ import unicode_literals

"""Common constants."""

KILOBYTE = 1024
"""Bytes in kilobyte."""

MEGABYTE = 1024 * KILOBYTE
"""Bytes in megabyte."""

GIGABYTE = 1024 * MEGABYTE
"""Bytes in gigabyte."""

TERABYTE = 1024 * GIGABYTE
"""Bytes in terabyte."""


MINUTE_SECONDS = 60
"""Seconds in minute."""

HOUR_SECONDS = 60 * MINUTE_SECONDS
"""Seconds in hour."""

DAY_SECONDS = 24 * HOUR_SECONDS
"""Seconds in day."""

WEEK_SECONDS = 7 * DAY_SECONDS
"""Seconds in week."""

MONTH_SECONDS = 30 * DAY_SECONDS
"""Seconds in average month."""


VOLUME_TYPE_THIN = "ThinProvisioned"
"""Volume type with thin provisioning."""

VOLUME_TYPE_THICK = "ThickProvisioned"
"""Volume type with think provisioning."""

VOLUME_TYPE_SNAPSHOT = "Snapshot"
"""Volume type Snapshot."""

VOLUME_TYPES = [
    VOLUME_TYPE_THIN,
    VOLUME_TYPE_THICK,
    VOLUME_TYPE_SNAPSHOT,
]
"""Valid volume types."""


VOLUME_REMOVE_ONLY_ME = "ONLY_ME"
"""Remove only current volume."""

VOLUME_REMOVE_DESCENDANTS = "INCLUDING_DESCENDANTS"
"""Remove current volume with descendants."""

VOLUME_REMOVE_DESCENDANTS_ONLY = "DESCENDANTS_ONLY"
"""Remove only descendants of current volume."""

VOLUME_REMOVE_VTREE = "WHOLE_VTREE"
"""Remove whole VTree."""

VOLUME_REMOVE_MODES = [
    VOLUME_REMOVE_ONLY_ME,
    VOLUME_REMOVE_DESCENDANTS,
    VOLUME_REMOVE_DESCENDANTS_ONLY,
    VOLUME_REMOVE_VTREE,
]
"""Valid volume remove modes"""



from __future__ import unicode_literals

from six import add_metaclass

from object_validator import validate, ValidationError
from object_validator import Integer, String, DictScheme

import pyscaleio.config
from pyscaleio import exceptions
from pyscaleio import utils


NETWORK_TIMEOUT = 30
"""Timeout for all network operations."""

REQUEST_RETRIES = 3
"""Default retries count for HTTP request."""

VOLUME_PREFIX = "/dev/disk/by-id"
"""Default prefix for volume path."""

VOLUME_NAME = "emc-vol-{system_id}-{volume_id}"
"""
Default name for volume path on SDC.
Must be parametrized with system_id and volume_id.
"""


@add_metaclass(utils.singleton)
class ScaleIOConfig(object):
    """ScaleIO config manager."""

    __scheme__ = {
        "network_timeout": Integer(min=0, optional=True),
        "request_retries": Integer(min=0, optional=True),
        "volume_prefix": String(optional=True),
        "volume_name": String(optional=True),
    }

    @classmethod
    def _get_scheme(cls):
        """Returns config scheme."""

        return DictScheme(cls.__scheme__)

    def _validate(self, options):
        """Validates config."""

        try:
            return validate("config", options, self._get_scheme())
        except ValidationError as e:
            raise exceptions.ScaleIOConfigError(e)

    def apply(self, **options):
        """Applies config options to a config module."""

        self._validate(options)

        for option, value in options.items():
            setattr(pyscaleio.config, option.upper(), value)



from __future__ import unicode_literals

import os
import pytest

from six import text_type as str

import pyscaleio
from pyscaleio import ScaleIOClient, ScaleIOClientsManager
from pyscaleio import StoragePool, Volume


TEST_NAME_PREFIX = "pyscaleiotest_"
"""Prefix for instance name."""


def _is_test_name(instance):
    """Returns True if the instance created by functonal test."""

    return instance.name and instance.name.startswith(TEST_NAME_PREFIX)


def _get_test_name(string):
    """Returns test name for resource."""

    return TEST_NAME_PREFIX + str(string)


@pytest.fixture(scope="module", autouse=True)
def client(request):
    """ScaleIO client fixture."""

    client = ScaleIOClient.from_args(
        os.getenv("scaleio_host"),
        os.getenv("scaleio_user"),
        os.getenv("scaleio_passwd"),
        os.getenv("scaleio_is_secure")
    )
    pyscaleio.add_client(client)
    request.addfinalizer(ScaleIOClientsManager().deregister)
    return client


def cleanup_volumes():
    """Removes all volumes from ScaleIO created by functional tests."""

    for volume in Volume.all():
        if not _is_test_name(volume):
            continue
        while volume.exports:
            volume.unexport()
            volume.update()
        volume.delete()

    assert not [v for v in Volume.all() if _is_test_name(v)]


def cleanup_pools():
    """Removes all storage pools from ScaleIO created by functional tests."""

    for pool in StoragePool.all():
        if not _is_test_name(pool):
            continue
        pool.delete()

    assert not [p for p in StoragePool.all() if _is_test_name(p)]


@pytest.fixture(autouse=True)
def setup_teardown(request):
    """Fixture that performs setup/teardown per test."""

    cleanup_volumes()
    cleanup_pools()
    request.addfinalizer(cleanup_volumes)
    request.addfinalizer(cleanup_pools)



from __future__ import unicode_literals

import pytest

from pyscaleio import ProtectionDomain, StoragePool, Volume
from pyscaleio import exceptions

from . import _get_test_name
from . import client, setup_teardown
assert client
assert setup_teardown


@pytest.fixture
def domain(request):
    return ProtectionDomain.all()[-1]


@pytest.fixture
def storage_pool(request):
    return StoragePool.all()[-1]


def test_create_storage_pool(domain):

    pool = StoragePool.create(domain["id"], name=_get_test_name(5))

    assert isinstance(pool, StoragePool)
    assert pool.name == _get_test_name(5)
    assert pool["protectionDomainId"] == domain["id"]

    assert pool.checksum_enabled is False
    assert pool.rfcache_enabled is False

    pool.delete()
    with pytest.raises(exceptions.ScaleIOError):
        pool.update()


def test_create_volume_in_storage_pool(domain, storage_pool):

    volume = storage_pool.create_volume(8, name=_get_test_name(1))
    assert isinstance(volume, Volume)
    assert volume.name == _get_test_name(1)

    assert volume["storagePoolId"] == storage_pool["id"]

    with pytest.raises(exceptions.ScaleIOError) as e:
        storage_pool.delete()
    assert "Storage Pool has volumes" in str(e.value)



from __future__ import unicode_literals

import collections
import json
import requests

import httmock
import mock
import pytest

from httmock import HTTMock
from psys import Error
from six import text_type as str

from pyscaleio import exceptions
from pyscaleio.client import ScaleIOSession, ScaleIOClient
from pyscaleio.manager import ScaleIOClientsManager
import pyscaleio.client
import pyscaleio.models


@pytest.fixture(scope="function")
def manager(request):
    m = ScaleIOClientsManager()
    request.addfinalizer(m.deregister)
    return m


@pytest.fixture(scope="function")
def mock_session(request):
    def generate_session(*args, **kwargs):
        if not args:
            args = ("localhost", "admin", "passwd")
        return ScaleIOSession(*args, **kwargs)
    return generate_session


@pytest.fixture(scope="function")
def mock_client(request):
    def generate_client(*args, **kwargs):
        if not args:
            args = ("localhost", "admin", "passwd")
        return ScaleIOClient.from_args(*args, **kwargs)
    return generate_client


@httmock.urlmatch(path=r".*login")
@httmock.remember_called
def login_payload(url, request):
    return httmock.response(200,
        json.dumps("some_random_token_string"),
        request=request
    )


@httmock.urlmatch(path=r".*logout")
@httmock.remember_called
def logout_payload(url, request):
    return httmock.response(200)


@httmock.urlmatch(path=r".*/api/version$", method="get")
def version_payload(url, request):
    return httmock.response(200, json.dumps(
        pyscaleio.client.__api_version__
    ), request=request)


@httmock.urlmatch(path=r".*/api/instances$", method="get")
def all_instances_payload(url, request):
    return httmock.response(200, {
        "resourceList": [1, 2],
    }, request=request)


def mock_instances_payload(resource, payload):
    path = r".*/api/types/{0}/instances".format(resource)

    @httmock.urlmatch(path=path, method="get")
    @httmock.remember_called
    def instances_of_payload(url, request):
        return httmock.response(200,
            json.dumps(payload), request=request)
    return instances_of_payload


@httmock.urlmatch(path=r".*/types/(\w+)/instances", method="post")
def create_instance_payload(url, request):
    return httmock.response(201, {"id": "test"}, request=request)


@pytest.mark.parametrize(("is_secure", "scheme"), [
    (True, "https"), (False, "http")
])
def test_session_initialize(mock_session, is_secure, scheme):

    client = mock_session(is_secure=is_secure)
    assert client.host == "localhost"
    assert client.user == "admin"
    assert client.passwd == "passwd"

    assert not client.token
    assert isinstance(
        client._ScaleIOSession__session,
        requests.Session)
    headers = client._ScaleIOSession__session.headers
    assert "Accept" in headers
    assert headers["Accept"] == "application/json; version=2.0"
    assert headers["Content-Type"] == "application/json"

    assert client.endpoint == "{0}://localhost/api/".format(scheme)


def test_session_login_positive(mock_session):

    client = mock_session()
    assert not client.token
    assert not client._ScaleIOSession__session.auth

    with HTTMock(login_payload):
        client.login()

    assert client.token == "some_random_token_string"
    assert client._ScaleIOSession__session.auth == ("admin", "some_random_token_string")


@pytest.mark.parametrize(("code", "message", "exc"), [
    (401, "Unauthorized", exceptions.ScaleIOAuthError),
    (500, "Server error", requests.HTTPError)
])
def test_session_login_negative(mock_session, code, message, exc):

    @httmock.urlmatch(path=r".*login")
    def login_payload(url, request):
        return httmock.response(code,
            json.dumps({
                "message": message,
                "httpStatusCode": code,
                "errorCode": 0
            }), request=request)

    client = mock_session()
    assert not client.token
    assert not client._ScaleIOSession__session.auth

    with HTTMock(login_payload):
        with pytest.raises(exc) as e:
            client.login()

    if isinstance(e, exceptions.ScaleIOError):
        assert e.status_code == code
        assert e.error_code == 0
        assert str(e) == message

    assert not client.token
    assert not client._ScaleIOSession__session.auth


def test_session_send_request(mock_session):

    @httmock.urlmatch(path=r"/api/test/instance")
    def request_payload(url, request):
        return httmock.response(200,
            json.dumps({"response": "test"}),
            request=request
        )

    client = mock_session()
    client.token = "some_token"

    with HTTMock(request_payload):
        result = client.get("test/instance")

    assert result == {"response": "test"}


@pytest.mark.parametrize(("effect", "result", "retries"), [
    (
        [
            (401, {"message": "Unauthorized", "httpStatusCode": 401}),
            (200, {"response": "test"})
        ],
        {"response": "test"}, 2
    ),
    (
        [
            (401, {"message": "Unauthorized", "httpStatusCode": 401}),
            (401, {"message": "Unauthorized", "httpStatusCode": 401}),
            (200, {"response": "test"})
        ],
        {"response": "test"}, 3
    )
])
def test_session_send_request_retries(mock_session, effect, result, retries):

    mock_handler = mock.Mock(side_effect=effect)

    @httmock.all_requests
    @httmock.remember_called
    def request_payload(url, request):
        code, payload = mock_handler()
        return httmock.response(code, json.dumps(payload),
            request=request
        )

    client = mock_session()
    client.token = "expired_token"

    with HTTMock(login_payload, request_payload):
        real_result = client.get("test/instance")

    assert real_result == result
    assert mock_handler.call_count == retries
    assert client.token == "some_random_token_string"

    assert login_payload.call["count"] == retries - 1
    assert request_payload.call["count"] == retries


def test_session_send_request_with_login(mock_session):

    @httmock.all_requests
    def api_payload(url, request):
        return httmock.response(200, json.dumps({"response": "test"}))

    client = mock_session()

    with HTTMock(login_payload, api_payload):
        result = client.get("test/instance")

    assert result == {"response": "test"}
    assert client.token == "some_random_token_string"


def test_session_send_request_negative(mock_session):

    @httmock.all_requests
    def api_exception(url, request):
        return httmock.response(500, json.dumps({
            "message": "Server error",
            "httpStatusCode": 500,
            "errorCode": 0
        }))

    client = mock_session()
    assert not client.token

    with HTTMock(login_payload, api_exception):
        with pytest.raises(exceptions.ScaleIOError) as e:
            client.get("test/instance")

    exc = e.value
    assert exc.status_code == 500
    assert exc.error_code == 0
    assert "code=500" in str(exc)
    assert "message=Server error" in str(exc)


def test_session_send_request_malformed(mock_session):

    @httmock.urlmatch(path=r"/api/test/instance")
    def request_payload(url, request):
        return httmock.response(200,
            "<?xml version='1.0' encoding='UTF-8'?>",
            request=request
        )

    client = mock_session()
    client.token = "some_token"

    with HTTMock(request_payload):
        with pytest.raises(exceptions.ScaleIOMalformedError):
            client.get("test/instance")


def test_session_logout(mock_session):

    client = mock_session()
    assert not client.token

    with HTTMock(logout_payload):
        client.logout()
    assert not logout_payload.call["called"]

    with HTTMock(login_payload, logout_payload):
        client.login()
        assert client.token

        client.logout()

    assert not client.token


def test_client_initialize(mock_session):

    with pytest.raises(Error) as e:
        ScaleIOClient(object())

    assert "must be initialized with ScaleIOSession" in str(e)

    client = ScaleIOClient(mock_session())
    assert client
    assert isinstance(client, ScaleIOClient)

    assert client.session
    assert isinstance(client.session, ScaleIOSession)

    with HTTMock(login_payload, version_payload):
        result = client.get_version()
        assert isinstance(result, float)
        assert result == pyscaleio.client.__api_version__


def test_client_getters(mock_client):

    client = mock_client()

    with HTTMock(login_payload, all_instances_payload):
        result = client.get_all_instances()
        assert isinstance(result, collections.MutableMapping)
        assert "resourceList" in result
        assert isinstance(result["resourceList"], list)

    instances_of_payload = mock_instances_payload(
        "Resource", [{"field": 1}, {"field": 2}]
    )
    with HTTMock(login_payload, instances_of_payload):
        result = client.get_instances_of("Resource")
        assert result
        assert isinstance(result, list)


def test_client_create_instance(mock_client):

    client = mock_client()

    with HTTMock(login_payload, create_instance_payload):
        instance_id = client.create_instance_of(
            "Volume", {"name": "test_volume"})
        assert isinstance(instance_id, str)
        assert instance_id == "test"


def test_client_lazy_get_system(mock_client):

    client = mock_client()

    systems_payload = mock_instances_payload(
        "System", [{"id": "test", "name": "test_name"}]
    )
    with HTTMock(login_payload, systems_payload):
        with mock.patch("pyscaleio.models.System.__scheme__", {}):
            result = client.system

            assert systems_payload.call["called"]
            assert systems_payload.call["count"] == 1

            assert isinstance(result, pyscaleio.models.System)
            assert result["id"] == "test"

            second_result = client.system
            assert result == second_result

            assert systems_payload.call["count"] == 1


def test_client_manager_register(manager):

    assert manager is ScaleIOClientsManager()

    client = ScaleIOClient.from_args("localhost", "admin", "passwd")
    manager.register(client)

    assert manager.clients
    assert len(manager.clients) == 1
    assert client.session.host in manager.clients

    assert manager.default is client

    registered_client = manager.get_client()
    assert registered_client is client

    manager.deregister("localhost")
    assert not manager.clients


def test_client_manager_register_negative(manager):

    assert manager is ScaleIOClientsManager()

    client = ScaleIOClient.from_args("localhost", "admin", "passwd")

    with pytest.raises(exceptions.ScaleIOEmptyClientRegistry):
        manager.get_client()

    with pytest.raises(exceptions.ScaleIOClientNotRegistered):
        manager.get_client("localhost")

    with pytest.raises(exceptions.ScaleIOInvalidClient):
        manager.register(object())

    client = ScaleIOClient.from_args("localhost", "admin", "passwd")
    manager.register(client)
    with pytest.raises(exceptions.ScaleIOClientAlreadyRegistered):
        manager.register(client)


def test_model_inject_client(manager, mock_client):

    manager.register(mock_client("localhost", "admin", "passwd"))

    assert pyscaleio.client._get_client({}) == pyscaleio.get_client()
    assert pyscaleio.client._get_client({"host": "localhost"}) == pyscaleio.get_client()

    second_client = mock_client("test_host", "admin", "passwd")
    assert pyscaleio.client._get_client({"client": second_client}) == second_client


def test_model_inject_client_negative():

    with pytest.raises(exceptions.ScaleIONotBothParameters):
        pyscaleio.client._get_client({"client": "test", "host": "test"})

    with pytest.raises(exceptions.ScaleIOInvalidClient):
        pyscaleio.client._get_client({"client": "test"})



from __future__ import unicode_literals

import pytest

from pyscaleio import System, StoragePool, Volume, Sdc
from pyscaleio import exceptions
from pyscaleio import constants

from . import _get_test_name
from . import client, setup_teardown
assert client
assert setup_teardown


@pytest.fixture
def storage_pool(request):
    return StoragePool.all()[-1]


@pytest.fixture
def system(request):
    return System.all()[-1]


def test_get_system():

    assert len(System.all()) == 1


def test_create_volume(storage_pool):

    volume = Volume.create(4, storage_pool["id"], name=_get_test_name(1))

    assert volume.name == _get_test_name(1)
    assert volume.size == 8 * constants.GIGABYTE

    volume.delete()
    with pytest.raises(exceptions.ScaleIOError):
        volume.update()

    # TODO: 'NotFound' error should be 404 instead of 500
    # assert e.value.status_code == 404
    # ScaleIOError: code=500, message=Could not find the volume


def test_volume_resize(storage_pool):

    volume = Volume.create(4, storage_pool["id"], name=_get_test_name(1))

    assert volume.size == 8 * constants.GIGABYTE

    with pytest.raises(exceptions.ScaleIOError) as e:
        volume.resize(10)
    assert e.value.status_code == 400

    with pytest.raises(exceptions.ScaleIOError) as e:
        volume.resize(6)
    assert e.value.status_code == 400

    volume.update()
    assert volume.size == 8 * constants.GIGABYTE

    volume.resize(16)
    volume.update()
    assert volume.size == 16 * constants.GIGABYTE


@pytest.mark.skipif("not Sdc.all()")
@pytest.mark.skipif("System.all()[-1].is_restricted and not Sdc.all_approved()")
def test_volume_export(storage_pool, system):

    volume = Volume.create(8, storage_pool["id"], name=_get_test_name(1))
    assert not volume.exports

    if system.is_restricted:
        sdc = Sdc.all_approved()[-1]
    else:
        sdc = Sdc.all()[-1]

    volume.export(sdc_id=sdc["id"])
    volume.update()

    assert volume.exports
    assert len(volume.exports) == 1
    assert volume.exports[0]["sdcId"] == sdc["id"]
    assert volume.exports[0]["sdcIp"] == sdc.ip

    with pytest.raises(exceptions.ScaleIOError):
        volume.export(sdc_id=sdc["id"])
        volume.export(sdc_guid=sdc["guid"])


@pytest.mark.skipif("not Sdc.all()")
@pytest.mark.skipif("System.all()[-1].is_restricted and not Sdc.all_approved()")
def test_volume_throttle(storage_pool, system):

    volume = Volume.create(8, storage_pool["id"], name=_get_test_name(1))
    assert not volume.exports

    if system.is_restricted:
        sdc = Sdc.all_approved()[-1]
    else:
        sdc = Sdc.all()[-1]

    volume.export(sdc_id=sdc["id"])
    volume.update()

    assert volume.exports
    assert len(volume.exports) == 1
    assert volume.exports[0]["limitIops"] == 0
    assert volume.exports[0]["limitBwInMbps"] == 0

    volume.throttle(sdc_id=sdc["id"], iops=1000)
    volume.update()

    assert volume.exports[0]["limitIops"] == 1000
    assert volume.exports[0]["limitBwInMbps"] == 0

    volume.throttle(sdc_id=sdc["id"], mbps=2048)
    volume.update()

    assert volume.exports[0]["limitIops"] == 1000
    assert volume.exports[0]["limitBwInMbps"] == 2048

    volume.throttle(sdc_id=sdc["id"], iops=2000, mbps=1024)
    volume.update()

    assert volume.exports[0]["limitIops"] == 2000
    assert volume.exports[0]["limitBwInMbps"] == 1024

    # FIXME: not working due to request validation error
    with pytest.raises(exceptions.ScaleIOError):
        volume.throttle(sdc_id=sdc["id"], mbps=0, iops=0)

    volume.throttle(sdc_id=sdc["id"], mbps=0)
    volume.update()
    assert volume.exports[0]["limitBwInMbps"] == 0


def test_create_volume_snapshot(storage_pool, system):

    volume = Volume.create(8, storage_pool["id"], name=_get_test_name(1))
    assert volume.size == 8 * constants.GIGABYTE

    snapshot = volume.snapshot(name=_get_test_name("snapshot_1"))
    assert isinstance(snapshot, Volume)

    assert snapshot.name == _get_test_name("snapshot_1")
    assert snapshot.type == constants.VOLUME_TYPE_SNAPSHOT
    assert snapshot["ancestorVolumeId"] == volume["id"]

    snapshot.delete()
    with pytest.raises(exceptions.ScaleIOError):
        snapshot.update()



from .client import ScaleIOSession, ScaleIOClient, inject  # noqa
from .config import ScaleIOConfig
from .manager import ScaleIOClientsManager
from .models import (
    System, ProtectionDomain, StoragePool,
    VTree, Sdc, Volume
)

__all__ = (
    ScaleIOSession.__name__, ScaleIOClient.__name__,
    System.__name__, ProtectionDomain.__name__,
    StoragePool.__name__, VTree.__name__, Sdc.__name__,
    Volume.__name__
)

__version__ = "0.1.3"

get_client = ScaleIOClientsManager().get_client
add_client = ScaleIOClientsManager().register
configure = ScaleIOConfig().apply



from __future__ import unicode_literals

import pytest
import mock

import pyscaleio
from pyscaleio import exceptions


def test_config_apply():

    assert pyscaleio.config.NETWORK_TIMEOUT == 30
    assert pyscaleio.config.REQUEST_RETRIES == 3
    assert pyscaleio.config.VOLUME_PREFIX == "/dev/disk/by-id"

    with mock.patch("pyscaleio.config", autospec=True):
        pyscaleio.configure(network_timeout=10)
        assert pyscaleio.config.NETWORK_TIMEOUT == 10

        pyscaleio.configure(volume_prefix="/dev")
        assert pyscaleio.config.VOLUME_PREFIX == "/dev"

        with pytest.raises(exceptions.ScaleIOConfigError):
            pyscaleio.configure(network_timeout="timeout")

        with pytest.raises(exceptions.ScaleIOConfigError):
            pyscaleio.configure(unexist_field="value")
